/*
 *  Copyright (C) 1999-2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Unhappily, we have to include this file in freshclam, because clamscan's
 * others.c is pthread-dependant. This one isn't.
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <pwd.h>
#include <errno.h>

#include "shared.h"
#include "others.h"
#include "defaults.h"
#include "md5.h"

#define bin_to_ascii(c) ((c)>=38?((c)-38+'a'):(c)>=12?((c)-12+'A'):(c)+'.')

char *md5_file(const char *filename)
{
	FILE *fd;
	unsigned char buffer[16];
	char *md5str;
	int i, cnt=0;

    if((fd = fopen(filename, "rb")) == NULL) {
	mprintf("!Can't read file in md5_file()\n");
	return NULL;
    }

    mprintf("*Calculating MD5(%s)... ", filename);
    md5_stream(fd, &buffer);
    md5str = (char*) calloc(32 + 1, sizeof(char));

    for(i=0; i<16; i++)
	cnt += sprintf(md5str + cnt, "%02x", buffer[i]);

    mprintf("*%s\n", md5str);

    return(md5str);
}

void mprintf(const char *str, ...)
{
	va_list args;
	FILE *fd;
	char logbuf[512];

    if(mprintf_disabled) {
	if(*str == '@') {
	    va_start(args, str);
	    vsnprintf(logbuf, 512, ++str, args);
	    va_end(args);
	    logg("ERROR: %s", logbuf);
	}
	return;
    }

    if(mprintf_stdout)
	fd = stdout;
    else
	fd = stderr;

/* legend:
 * ! - error
 * @ - error with logging
 * ...
 */

/*
 *             ERROR    WARNING    STANDARD
 * normal       yes       yes        yes
 * 
 * verbose      yes       yes        yes
 * 
 * quiet        yes       no         no
 */

    va_start(args, str);

    if(*str == '!') {
	fprintf(fd, "ERROR: ");
	vfprintf(fd, ++str, args);
    } else if(*str == '@') {
	fprintf(fd, "ERROR: ");
	vfprintf(fd, ++str, args);
	vsnprintf(logbuf, 512, str, args);
	logg("ERROR: %s", logbuf);
    } else if(!mprintf_quiet) {
	if(*str == '^') {
	    fprintf(fd, "WARNING: ");
	    vfprintf(fd, ++str, args);
	} else if(*str == '*') {
	    if(mprintf_verbose)
		vfprintf(fd, ++str, args);
	} else vfprintf(fd, str, args);
    }

    va_end(args);

    if(fd == stdout)
	fflush(stdout);

}

int logg(const char *str, ...)
{
	va_list args;
	static FILE *fd = NULL;

    if(logfile) {

	if(!fd) {
	    if((fd = fopen(logfile, "a")) == NULL) {
		mprintf("!LOGGER: Can't open file %s to write.\n", logfile);
		return 1;
	    }
	}

	va_start(args, str);

	if(*str == '!') {
	    fprintf(fd, "ERROR: ");
	    vfprintf(fd, ++str, args);
	} else if(*str == '^') {
	    fprintf(fd, "WARNING: ");
	    vfprintf(fd, ++str, args);
	} else if(*str == '*') {
	    if(logverbose)
		vfprintf(fd, ++str, args);
	} else vfprintf(fd, str, args);

	va_end(args);

    }

    fflush(fd);
    return 0;
}

int *gensalt(void)
{
	int *salt;
	int i;

    salt = (int *) mmalloc(16 * sizeof(int));

    for(i = 0; i < 16; i++)
	salt[i] = rndnum(255);

    return salt;
}

char *gentemp(const char *dir)
{
	char *name, *mdir;
	unsigned char buffer[16];
	int cnt=0, i;
	struct md5_ctx ctx;

    if(!dir)
	mdir = "/tmp";
    else
	mdir = (char *) dir;

    name = (char*) calloc(strlen(mdir) + 1 + 16 + 1, sizeof(char));
    cnt += sprintf(name, "%s/", mdir);

    do {
	md5_init_ctx(&ctx);
	md5_process_bytes(gensalt(), 16, &ctx);
	md5_finish_ctx(&ctx, &buffer);

	for(i=0; i<8; i++)
	    cnt += sprintf(name + cnt, "%02x", buffer[i]);

	mprintf("*Generated temporary name %s\n", name);
    } while(info(name, 1) != -1);

    return(name);
}

void *mcalloc(size_t nmemb, size_t size)
{
	void *alloc;

    alloc = calloc(nmemb, size);

    if(!alloc) {
	mprintf("!CRITICAL: Can't allocate memory.\n");
	perror("calloc_problem");
	mprintf("exit() forced...\n");
	exit(70);
	return NULL;
    } else return alloc;
}

void *mmalloc(size_t size)
{
	void *alloc;

    alloc = malloc(size);

    if(!alloc) {
	mprintf("!CRITICAL: Can't allocate memory.\n");
	perror("malloc_problem");
	mprintf("exit() forced...\n");
	exit(71);
	return NULL;
    } else return alloc;
}

int strbcasestr(const char *haystack, const char *needle)
{
	char *pt = (char *) haystack;
	int i, j;

    i = strlen(haystack);
    j = strlen(needle);

    if(i < j)
	return 0;

    pt += i - j;

    return !strcasecmp(pt, needle);
}

int info(const char *filename, int i)
{
	struct stat infostruct;

    if(stat(filename, &infostruct) == -1)
	return(-1);

    switch(i) {

	case 1: /* size */
	    return infostruct.st_size;
	case 2: /* permissions */
	    return (mode_t)infostruct.st_mode;
	case 3: /* modification time */
	    return infostruct.st_mtime;
	case 4: /* UID */
	    return infostruct.st_uid;
	case 5: /* GID */
	    return infostruct.st_gid;
	default:
	    mprintf("!info(): Unknown option\n");
	    exit(1);
    }
}

/* this functions returns pseudo random number between [0, max) */

#ifndef C_URANDOM
/* it's very weak */

int rndnum(int max)
{
    struct timeval tv;

  gettimeofday(&tv, (struct timezone *) 0);
  srand(tv.tv_usec+clock());

  return rand() % max;
}

#else

int rndnum(int max)
{
	static FILE *fd = NULL;
	unsigned int generated;
	char *byte;
	int size;

    if(fd == NULL) {
	mprintf("*Opening /dev/urandom.\n");
	if((fd = fopen("/dev/urandom", "rb")) == NULL) {
	    mprintf("!Can't open /dev/urandom.\n");
	    return -1;
	}
    }

    byte = (char *) &generated;
    size = sizeof(generated);
    do {
	int bread;
	bread = fread(byte, 1, size, fd);
	size -= bread;
	byte += bread;
    } while(size > 0);

    return generated % max;
}

#endif
